/*
    Copyright 2005,2006,2007 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <ctype.h>

#ifdef __MINGW32__
    #define NTS     "I64"
#else
    #define NTS     "ll"
#endif



#define VER         "0.2a"
#define MAXFMT      8
#define DECHEX(S,V) (toupper(S[1]) == 'X') ? sscanf(S, "%x", &V) : sscanf(S, "%u", &V)
#define SHOW(x)     fprintf(stdout, cfmt, *(x *)out)
#define FMTHELP     " fmt is the number of bytes that compose each element:\n" \
                    "   1 for char (8 bits), 2 for short (16), 4 for int (32) and 8 for 64 bits\n"



typedef uint64_t (*bswap_proto)(uint64_t value);
uint64_t noswap(uint64_t value);
uint64_t bswap8(uint64_t value);
uint64_t bswap16(uint64_t value);
uint64_t bswap32(uint64_t value);
uint64_t bswap64(uint64_t value);
int get_num(char *str);
void varbytes(int show);
int findfloat(char *p);
int find_endian(void);
void std_err(void);



int main(int argc, char *argv[]) {
    FILE        *fd;
    uint64_t    value   = 0,
                and     = 0;
    int         off     = 0,
                size    = 0xffffffff;
    int         i,
                lf,
                mlf     = 0,
                fmt,
                space   = 0,
                hex     = 1,
                endian  = 0,
                bigefix = 0,
                showvar = 0,
                skip    = 0,
                myfloat;
    char        *cfmt  = NULL,
                *fmts  = "";
    uint8_t     out[MAXFMT + 4];    // more compatibility?
    bswap_proto swap;

    setbuf(stdin, NULL);
    setbuf(stdout, NULL);
    setbuf(stderr, NULL);

    fputs("\n"
        "Byte2C "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stderr);

    if(argc < 3) {
        printf("\n"
            "Usage: %s [options] <file> <fmt>\n"
            "\n"
            "Options:\n"
            "-o OFF   offset from which starting the dump\n"
            "-s NUM   amount of elements to handle\n"
            "-k NUM   number of bytes in the file to skip after each element\n"
            "-l NUM   number of elements to show in each line (automatic)\n"
            "-p NUM   number of spaces to add at the left of each visualized element\n"
            "-d       shortcut for decimal output (default is an automatic hex, look at -f)\n"
            "-f STR   manual handling of the format used to visualize the element, like\n"
            "         \"0x%%02x, \", \"%%d \", \" %%08x \", \" %%"NTS"x \", %%02x, \"char %%c\" and so on\n"
            "-e       enable big endian/network byte order elements (default little endian)\n"
            "-i       show also the size of the C variables on your system\n"
            "\n"
            FMTHELP
            "\n", argv[0]);
        exit(1);
    }

    argc -= 2;
    for(i = 1; i < argc; i++) {
        switch(argv[i][1]) {
            case 'o': off     = get_num(argv[++i]);     break;
            case 's': size    = get_num(argv[++i]);     break;
            case 'k': skip    = get_num(argv[++i]);     break;
            case 'l': mlf     = get_num(argv[++i]);     break;
            case 'p': space   = get_num(argv[++i]);     break;
            case 'd': hex     = 0;                      break;
            case 'f': cfmt    = argv[++i];              break;
            case 'e': endian  = 1;                      break;
            case 'i': showvar = 1;                      break;
            default: {
                printf("\nError: wrong command-line argument (%s)\n\n", argv[i]);
                exit(1);
                }
        }
    }

    if(!strcmp(argv[argc], "-")) {
        fd = stdin;
    } else {
        fd = fopen(argv[argc], "rb");
        if(!fd) std_err();
    }

    fmt = atoi(argv[argc + 1]);
    if((fmt < 1) || (fmt > MAXFMT)) {
        printf("\nError: the maximum format size is %d\n\n", MAXFMT);
        exit(1);
    }
    if((fmt & 1) && (fmt != 1)) {
        printf("\nError: wrong format size\n"
            FMTHELP"\n");
        exit(1);
    }

    if(off) {
        if(fd == stdin) {
            for(i = 0; i < off; i++) {
                if(fgetc(fd) < 0) std_err();
            }
        } else {
            if(fseek(fd, off, SEEK_SET) < 0) std_err();
        }
    }

    if(!mlf) {
        mlf = 16 / fmt;
        if(!mlf) mlf++;
    }

    switch(fmt) {
        case 1: fmts = "hh";    break;
        case 2: fmts = "h";     break;
        case 4: fmts = "";      break;
        case 8: fmts = NTS;     break;
    }

    if(!cfmt) {
        if(hex) {
            cfmt = malloc(16);
            sprintf(cfmt, "0x%%0%d%sx, ", fmt << 1, fmts);
        } else {
            cfmt = "%u, ";
        }
    }

    myfloat = findfloat(cfmt);

    and = (1 << (fmt << 3)) - 1;
    if(!and) and--;

    swap = noswap;
    if(endian) {
        switch(fmt) {
            case 1: swap = bswap8;  break;
            case 2: swap = bswap16; break;
            case 4: swap = bswap32; break;
            case 8: swap = bswap64; break;
        }
    }

    if(find_endian()) bigefix = (MAXFMT - fmt) << 3;

    varbytes(showvar);

    for(lf = mlf; fread(&value, fmt, 1, fd) && size--; lf++) {
        if(lf == mlf) {
            fputc('\n', stdout);
            for(i = 0; i < space; i++) fputc(' ', stdout);
            lf = 0;
        }

        value = (swap(value >> bigefix) & and); // required for
        memset(out, 0, sizeof(out));            // maximum
        for(i = 0; i < fmt; i++) {              // compatibility
            out[i] = (value >> (i << 3)) & 0xff;
        }

        if(myfloat == 1) {
            switch(fmt) {
                case 1: SHOW(float);        break;
                case 2: SHOW(float);        break;
                case 4: SHOW(float);        break;
                case 8: SHOW(float);        break;
            }
        } else if(myfloat == 2) {
            switch(fmt) {
                case 1: SHOW(double);       break;
                case 2: SHOW(double);       break;
                case 4: SHOW(double);       break;
                case 8: SHOW(double);       break;
            }
        } else {
            switch(fmt) {
                case 1: SHOW(uint8_t);      break;
                case 2: SHOW(uint16_t);     break;
                case 4: SHOW(uint32_t);     break;
                case 8: SHOW(uint64_t);     break;
            }
        }

        if(skip) {
            if(fd == stdin) {
                for(i = 0; i < skip; i++) {
                    if(fgetc(fd) < 0) std_err();
                }
            } else {
                if(fseek(fd, skip, SEEK_CUR) < 0) std_err();
            }
        }
    }

    fputc('\n', stdout);
    if(fd != stdin) fclose(fd);
    return(0);
}



uint64_t noswap(uint64_t value) {
    return(value);
}



uint64_t bswap8(uint64_t value) {
    return(value & 0xff);
}



uint64_t bswap16(uint64_t value) {
    return((value & 0xff00) >> 8) |
          ((value & 0x00ff) << 8);
}



uint64_t bswap32(uint64_t value) {
    return((value & 0xff000000) >> 24) |
          ((value & 0x00ff0000) >>  8) |
          ((value & 0x0000ff00) <<  8) |
          ((value & 0x000000ff) << 24);
}



uint64_t bswap64(uint64_t value) {
    uint32_t    tmp1,
                tmp2;

    tmp1 = value >> 32;
    tmp2 = value & 0xffffffff;

    tmp1 =  ((tmp1 & 0xff000000) >> 24) |
            ((tmp1 & 0x00ff0000) >>  8) |
            ((tmp1 & 0x0000ff00) <<  8) |
            ((tmp1 & 0x000000ff) << 24);
    tmp2 =  ((tmp2 & 0xff000000) >> 24) |
            ((tmp2 & 0x00ff0000) >>  8) |
            ((tmp2 & 0x0000ff00) <<  8) |
            ((tmp2 & 0x000000ff) << 24);

    value = tmp2;
    return((value << 32) | tmp1);
/*
    return((value & 0xff00000000000000UL) >> 56) |
          ((value & 0x00ff000000000000UL) >> 40) |
          ((value & 0x0000ff0000000000UL) >> 24) |
          ((value & 0x000000ff00000000UL) >>  8) |
          ((value & 0x00000000ff000000UL) <<  8) |
          ((value & 0x0000000000ff0000UL) << 24) |
          ((value & 0x000000000000ff00UL) << 40) |
          ((value & 0x00000000000000ffUL) << 56);
*/
}



int get_num(char *str) {
    int     offset;

    if(!strncmp(str, "0x", 2) || !strncmp(str, "0X", 2)) {
        sscanf(str + 2, "%x", &offset);
    } else {
        sscanf(str, "%u", &offset);
    }
    return(offset);
}



void varbytes(int show) {
    int     charsz,
            shortsz,
            intsz,
            longsz,
            llongsz,
            floatsz,
            doublesz,
            ldoublesz,
            unsignedsz;

    charsz     = sizeof(char)        * 8;
    shortsz    = sizeof(short)       * 8;
    intsz      = sizeof(int)         * 8;
    longsz     = sizeof(long)        * 8;
    llongsz    = sizeof(long long)   * 8;
    floatsz    = sizeof(float)       * 8;
    doublesz   = sizeof(double)      * 8;
    ldoublesz  = sizeof(long double) * 8;
    unsignedsz = sizeof(unsigned)    * 8;

    if(show) {
        printf(
            "  char:        %d\n"
            "  short:       %d\n"
            "  int:         %d\n"
            "  long:        %d\n"
            "  long long:   %d\n"
            "  float:       %d\n"
            "  double:      %d\n"
            "  long double: %d\n"
            "  unsigned:    %d\n",
            charsz,
            shortsz,
            intsz,
            longsz,
            llongsz,
            floatsz,
            doublesz,
            ldoublesz,
            unsignedsz);
    }
}



int findfloat(char *p) {
    int     next;

    for(; *p; p++) {
        if((p[0] == '%') && (p[1] != '%')) {
            for(++p; *p; p++) {
                switch(*p){
                    case '.':
                    case '#':
                    case '-':
                    case ' ':
                    case '+':
                    case '\'':
                    case 'I':
                    case '0':
                    case '1':
                    case '2':
                    case '3':
                    case '4':
                    case '5':
                    case '6':
                    case '7':
                    case '8':
                    case '9':
                    case 'h':
                    case 'l':
                    case 'L':
                    case 'q':
                    case 'j':
                    case 'z':
                    case 't':
                        next = 1;   break;
                    case 'f':
                    case 'F':
                        return(1);  break;
                    case 'e':
                    case 'E':
                    case 'g':
                    case 'G':
                    case 'a':
                    case 'A':
                        return(2);  break;
                    default:
                        next = 0;   break;
                }
                if(!next) break;
            }
        }
    }
    return(0);
}



int find_endian(void) { // 0 = little, 1 = big endian
    int     endian = 1;

    if(*(char *)&endian) endian = 0;
    return(endian);
}



void std_err(void) {
    perror("\nError");
    exit(1);
}


